
# $Id$

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import datetime, sys, subprocess, re, traceback, time
from grid_monitoring.common import HostInfo, QueueInfo, JobInfo, ReservationInfo
from grid_monitoring.gmexceptions import GMInternalError, GMExternalError, GMUnknowJob

def _runCommand (command):
    try:
        sr = subprocess.Popen (command, stdout=subprocess.PIPE, shell=True)
        commandOutput = sr.communicate()[0]
    except OSError as e:
        raise GMExternalError (command + " command failed, reason:" + str (e))

    if sr.returncode != 0:
        raise GMExternalError (command + " command returned code: " + str (sr.returncode) + ". Details: " + commandOutput)

    return commandOutput


def _output2dicts (output, keys):
    '''Convert lists separated by an exclamation point (!) to dicts with given keys'''

    if re.search ("There is currently no .* status to report.", output):
        return []

    return [dict (zip (keys, line.split('!'))) for line in output.splitlines()]


def _str2datatime (s):
    '''Convert string in proper format to datetime object'''

    return datetime.datetime (*(time.strptime (s, "%m/%d/%Y %H:%M")[0:6]))


def _str2datatime2 (s):
    '''Convert string in proper format to datetime object'''

    return datetime.datetime (*(time.strptime (s, "%a %d %b %Y %I:%M:%S %p %Z")[0:6]))


def getHostData (hostnames=None):
    statusOutput = _runCommand ("llstatus -r %n %cpu %l %v %d %m %sts %a %o")
    nodeList = _output2dicts (statusOutput, ['name', 'cpus', 'ldAvg', 'freeVMem',
                                             'freeDisk', 'mem', 'state', 'arch', 'opSys'])

    jobs = getJobData()
    nodeJobs = {}
    for job in jobs:
        for node in job.ExecutionHosts.iterkeys():
            if node not in nodeJobs:
                nodeJobs[node] = []

            nodeJobs[node].append (job.JobID)

    hostInfoList = []
    for node in nodeList:
        hostInfo = HostInfo()
        hostInfo.HostName = node['name']
        hostInfo.CpuCount = int (node['cpus'])
        hostInfo.CpuUtilization = float (node['ldAvg'])
        hostInfo.FreeSwap = float (node['freeVMem']) * 2**10 # kB to B
        hostInfo.FreeFilesystem = float (node['freeDisk']) * 2**10 # kB to B
        hostInfo.TotalMemory = float (node['mem']) * 2**10 # kB to B

        if node['state'] in ["Busy", "Idle", "Running"]:
            hostInfo.State = "up"
        elif node['state'] in ["Down"]:
            hostInfo.State = "down"
        elif node['state'] in ["Drained", "Draining", "Flush", "None", "Suspend"]:
            hostInfo.State = "offline"

        # default powerpc, node['arch'] not used
        hostInfo.CpuArch = "powerpc"

        # default linux, node['opSys'] not used
        hostInfo.OSName = "linux"

        if hostInfo.HostName in nodeJobs:
            hostInfo.Jobs = nodeJobs[hostInfo.HostName]

        #TODO ?
        #hostInfo.Queue = (llstatus -r )

        hostInfoList.append (hostInfo)

    return hostInfoList


def getJobData (jobids=None):
    llqOutput = _runCommand ("llq -r %id %jn %o %st %c %R %dq %dd %dc %h")
    jobList = _output2dicts (llqOutput, ['id', 'name', 'owner', 'status', 'class', 'resId',
                                         'subTime', 'startTime', 'endTime', 'nodes'])

    jobInfoList = []

    for job in jobList:
        jobInfo = JobInfo()
        jobInfo.JobID = job['id']
        jobInfo.JobName = job['name']
        jobInfo.OwnerName = job['owner']
        jobInfo.QueueName = job['class']
        #jobInfo.Description = (llq -l, comment)

        if job['status'] in ['D', 'I', 'NQ', 'NR', 'P', 'X', 'XP', 'RP', 'ST', 'S', 'HS', 'H']:
            jobInfo.JobStatus = 'Pending'
        elif job['status'] in ['CK', 'CP', 'EP', 'MP', 'R']:
            jobInfo.JobStatus = 'Running'
        elif job['status'] in ['E', 'V', 'VP']:
            jobInfo.JobStatus = 'Suspended'
        elif job['status'] in ['CA', 'C', 'RM', 'TX']:
            jobInfo.JobStatus = 'Done'
        else:
            jobInfo.JobStatus = 'Unknown'

        if len (job['resId']) > 0:
            jobInfo.ReservationID = job['resId']

        jobInfo.SubmissionTime = _str2datatime (job['subTime'])

        if len (job['startTime']) > 0:
            jobInfo.StartTime = _str2datatime (job['startTime'])

        if len (job['endTime']) > 0:
            jobInfo.EndTime = _str2datatime (job['endTime'])

        #FIXME precise number
        if len (job['nodes']) > 0:
            for node in job['nodes'].split(','):
                jobInfo.ExecutionHosts[node] = 1

        jobInfoList.append (jobInfo)

    return jobInfoList


def getQueueData (queuenames=None):
    llclassOutput = _runCommand ("llclass -l")

    classDict = {}
    cl = None
    for line in llclassOutput.splitlines():
        match = re.search ("^=+ Class (?P<name>\w+) =+$", line)

        if match is not None:
            cl = match.group('name')
            classDict[cl] = {}
            classDict[cl]['PendingJobs'] = []
            classDict[cl]['RunningJobs'] = []
        elif re.search ("^-+$", line):
            break
        elif cl is not None:
            prop, val = line.split (':', 1)
            classDict[cl][str.strip (prop)] = str.strip (val)

    jobInfoList = getJobData()
    for job in jobInfoList:
        if job.QueueName in classDict:
            if job.JobStatus == 'Pending':
                classDict[job.QueueName]['PendingJobs'].append (job.JobID)
            elif job.JobStatus == 'Running':
                classDict[job.QueueName]['RunningJobs'].append (job.JobID)

    queueInfoList = []

    for cl in classDict:
        clProps = classDict[cl]
        queueInfo = QueueInfo()
        queueInfo.QueueName = cl
        queueInfo.QueuePriority = int (clProps['Priority'])
        queueInfo.QueueEnabled = True
        queueInfo.MaxJobs = int (clProps['Maximum_slots'])
        queueInfo.PendingJobs = len (clProps['PendingJobs'])
        queueInfo.RunningJobs = len (clProps['RunningJobs'])
        queueInfo.JobIDs = clProps['PendingJobs'] + clProps['RunningJobs']
        queueInfo.IsDefault = bool (cl == 'No_Class')

        match = re.search ("\((?P<name>\d+) seconds", clProps['Wall_clock_limit'])

        if match is not None:
            queueInfo.Walltime = datetime.timedelta (seconds = int (match.group('name')))

        queueInfoList.append (queueInfo)

    return queueInfoList


def getReservationData (reservation_names=None):
    llqresOutput = _runCommand ("llqres -r")
    resList = _output2dicts (llqresOutput, ['id', 'creationTime', 'owner', 'group',
                                            'startTime', 'duration', 'endTime', 'shared',
                                            'rmOnIdle', 'status', 'modBy', 'modeTime',
                                            'usersNo', 'usersList', 'groupsNo', 'groupsList',
                                            'nodesNo', 'nodesList', 'jobsNo', 'jobsList'])

    reservationInfoList = []

    for res in resList:
        resInfo = ReservationInfo()
        resInfo.ReservationName = res['id']
        resInfo.Type = 'ar'
        resInfo.StartTime = _str2datatime2 (res['startTime'])
        resInfo.EndTime = _str2datatime2 (res['endTime'])
        resInfo.Owner = res['owner']
        #resInfo.Resources =

        reservationInfoList.append (resInfo)

    return reservationInfoList
