
# $Id$

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

try:
    from qcg.core import *
except ImportError:
    def error(msg):
        sys.stderr.write("[ERROR] " + msg)
    pass #for testing only

import datetime, sys, subprocess, re, traceback, time, os
from grid_monitoring.common import HostInfo, QueueInfo, JobInfo, ReservationInfo, asList, asDict
from grid_monitoring.gmexceptions import GMInternalError, GMExternalError, GMUnknowJob
try:
    import libxml2 
except ImportError:
    error("Unable to load libxml2. No reservation info woulbe be returned")


__jobsRegex = re.compile(r'[, \d]+/([^,]*)(.*)')
#__statusRegex = re.compile(r'([^=]+)=([^,]*),*')
__statusRegex = re.compile(r'([^=]+)=(.*?)($|(,(?=[a-zA-Z_0-9]+?=)))')
__queueStateRegex = re.compile(r'([^:]+):([^ ]*)[ ]*')
__execHostRegex = re.compile(r'([^/]+)/\d+\+*')
__nodesRegex = re.compile(r'(\d+)(:ppn=(\d+))?[^+]*\+?')
__valueRegex = re.compile(r'([\d\.]+)(\D*)')
__walltimeRegex = re.compile(r'([0-9]+):([0-9]+):([0-9]+)')
__resACLRegex = re.compile(r'^\s*ACL:.*USER==(\w+)', re.MULTILINE)
__resIdRegex = re.compile(r'FLAGS:ADVRES:(\S+)', re.MULTILINE)
__resAccountingRegex = re.compile(r'^\s*Accounting.*User:(\w+)', re.MULTILINE)
__newlineRegex = re.compile(r'\n')
__emptyLineRegex = re.compile(r'\n[\s\n]*\n')
__equalRegex = re.compile(r'[ ]+=[ ]+')
__jobIdRegex = re.compile(r'[^\d]*(\d+.*)')
__queueRegex = re.compile(r'Queue: (.+)')
__showstartRegex = re.compile(r'^[\w\s]+start in\s+(([0-9]+):)?([0-9]+):([0-9]+):([0-9]+)', re.MULTILINE) # 13:02:08:33 or 00:00:00
__showstartInfRegex = re.compile(r'^[\w\s]+start in\s+INFINITY', re.MULTILINE) # 13:02:08:33 or 00:00:00
__hostsInfoDict = None
__hostsInfoList = None
__jobsInfoDict = None
__jobsInfoList = None
__queuesInfoList = None
__queuesInfoDict = None
__reservationsInfoDict = None
__reservationsInfoList = None

def __checknodeResRegexExp(value):
    return "User \'" + value + "\'.*Procs: (\d+)/(\d+)" 

__defaultOwner = "root"
__checknodeROI = "Reservations:"

__postfixTable = {
    "": 1,
    "b": 1,
    "kb": 1024,
    "mb": 1024*1024,
    "gb": 1024*1024*1024
}

def __extractPostFix(value):
    m = __valueRegex.match(value)
    ret = [m.group(1)]
    if m.group(2)!=None:
        ret.append(m.group(2))
    else:
        ret.append("")
    return ret

def __calculateValue(value):
    try:
      p = __extractPostFix(value)
      return float(round(float(p[0]) \
        *__postfixTable.get(p[1]))) 
    except Exception:
      return float(0)


def __parseFloatValue(value):
    try:
      return float(value)
    except Exception:
      return float(0)


def __commandOutputToDict(output, items_list=None, key_regex=None):
    blocks_list = __emptyLineRegex.split(output)
    items_dictionary = {}
    for block in blocks_list:
        if block.lstrip() == "":
            continue
        block_lines = __newlineRegex.split(block)

        if key_regex != None:
          block_name = key_regex.match(block_lines[0]).group(1)
        else:
          block_name = block_lines[0]

        block_lines = block_lines[1:]
        block_dict = { }
        for line in block_lines:
            tuple = __equalRegex.split(line, 1)
            if len(tuple) == 2:
              block_dict[ tuple[0].strip() ] = tuple[1].strip()

        if items_list != None:
          items_list.append(block_name)

        items_dictionary[block_name] = block_dict

    return items_dictionary

def getHostData():
    global __hostsInfoDict
    global __hostsInfoList
    if __hostsInfoList is not None:
        return (__hostsInfoList, __hostsInfoDict)

    __hostsInfoDict = dict()
    __hostsInfoList = []
    pbsnodes_output = _pbsnodes(['-a'] )
    nodes_filter = os.getenv("QCG_NODE_FILTER")
    if nodes_filter != None:
        try:
            nodes_filter_key = nodes_filter.split(":")[0]
            nodes_filter_regex_str = nodes_filter.split(":")[1]
            if nodes_filter_regex_str[:1] == '~':
                nodes_filter_negative = True
                nodes_filter_regex_str = nodes_filter_regex_str[1:]
            else:
                nodes_filter_negative = False
            nodes_filter_regex = re.compile(nodes_filter_regex_str)
        except Exception:
            raise GMInternalError("QCG_NODE_FILTER: Invalid format: '" + nodes_filter + "':. Expected: 'field:regex'")

    pbsnodes_nodes_list = __emptyLineRegex.split(pbsnodes_output)
    nodes_list = []
    nodes_dictionary = __commandOutputToDict(pbsnodes_output, nodes_list)

    for node_name in nodes_list:
        current_host = HostInfo()
        current_host.HostName = node_name;
        node_properties = nodes_dictionary[node_name];
        if nodes_filter != None:
            if nodes_filter_key not in node_properties:
                current_host.Dedicated = True
            if (nodes_filter_regex.search(node_properties[nodes_filter_key]) == None) ^ nodes_filter_negative:
                current_host.Dedicated = True
        if node_properties['state'].find("down") != -1: 
          current_host.State = "down"
        elif node_properties['state'].find("offline") != -1:
          current_host.State = "offline"
        else:
          current_host.State = "up"
        current_host.CpuCount = int(node_properties['np'])

        #jobs
        if 'jobs' in node_properties:
            m = __jobsRegex.match(node_properties['jobs'])
            current_host.Jobs = []
            while m!=None:
              current_host.Jobs.append(m.group(1))
              jobs = m.group(2)
              m = __jobsRegex.match(jobs)

            current_host.UsedCpuCount = len(current_host.Jobs)

        if 'properties' in node_properties:
          current_host.AdditionalResources = node_properties['properties']

        if 'status' in node_properties:
          #decode status
          attributes = __decodeHostStatusField(node_properties['status'])

          if attributes.get('opsys') != None:
            current_host.OSName = attributes["opsys"]

          if attributes.get('uname') != None:
            uname = attributes["uname"]
            if uname.find("x86_64") != -1:
              current_host.CpuArch = "x86_64"
            elif uname.find("x86") != -1:
              current_host.CpuArch = "x86"
            elif uname.find("IA-64") != -1:
              current_host.CpuArch = "IA-64"

          if current_host.CpuArch == "Unknown" and properties != None:
            if properties.find("x86_64") != -1:
              current_host.CpuArch = "x86_64"
            elif properties.find("x86") != -1:
              current_host.CpuArch = "x86"
            elif properties.find("IA-64") != -1:
              current_host.CpuArch = "IA-64"

          if attributes.get('physmem') != None:
            current_host.TotalMemory = __calculateValue(attributes["physmem"])

          if attributes.get('totmem') != None:
            current_host.TotalSwap = __calculateValue(attributes["totmem"]) - current_host.TotalMemory

          if attributes.get('loadave') != None:
              current_host.CpuUtilization = __parseFloatValue((attributes["loadave"]))

        if 'gpus' in node_properties:
            gpus = int(node_properties['gpus'])
            if gpus > 0:
                current_host.GpuCount = gpus

        __hostsInfoDict[node_name]=current_host
        __hostsInfoList.append(current_host)

    return (__hostsInfoList,__hostsInfoDict)

def __decodeHostStatusField(status):
    ret = {} 
    for m in __statusRegex.finditer(status):
        ret[m.group(1)] = m.group(2)        
    return ret

def __decodeQueueStateField(state):
    ret = {}
    for m in __queueStateRegex.finditer(state):
        ret[m.group(1)] = m.group(2)
    return ret

def __decodeTimeDelta(time):
    match = __walltimeRegex.match(time)
    sec = int(match.group(1)) * 3600 + int(match.group(2)) * 60 + int(match.group(3))
    return datetime.timedelta(seconds=sec) 

def __decodeNodes(nodes):
    totalCores = 0
    try:
        for m in __nodesRegex.finditer(nodes):
            if m.group(3) is not None:
                totalCores = totalCores + int(m.group(3)) * int(m.group(1)) #nodes + ppn
            else:
                totalCores = totalCores + int(m.group(1)) #only nodes
        
        if totalCores != 0:
            return totalCores
        else:
            return None
    except:
        error("Failed to parse: nodes: " + nodes)
        return None

def __decodeShowstartTimeDelta(time):
    match = __showstartRegex.search(time)
    if match is None:
        if __showstartInfRegex.search(time) is None:
            error("failed to match showstart output: " + time)
        else:
            return datetime.timedelta(days=1024) #1024 days = infinity
        return None
    
    sec = int(match.group(3)) * 3600 + int(match.group(4)) * 60 + int(match.group(5))
    if match.group(2) is not None:
        sec += int(match.group(2)) * 3600 * 24
    return datetime.timedelta(seconds=sec)
        
def __decodeHostData(info):
    current_host = HostInfo()
    current_host.HostName = info[0]
    map = info[1]
    if "status" in map.keys():
        attributes = __decodeHostStatusField(info[1]["status"])
    else:
        return current_host
    
    if "properties" in map.keys():
        properties = info[1]["properties"]
    else:
        properties = None
            
    current_host.OSName = attributes["opsys"]   
    
    uname = attributes["uname"]
    if uname.find("x86_64") != -1:
        current_host.CpuArch = "x86_64"
    elif uname.find("x86") != -1:
        current_host.CpuArch = "x86"
    elif uname.find("IA-64") != -1:
        current_host.CpuArch = "IA-64"      

    if current_host.CpuArch == "Unknown" and properties != None:        
        if properties.find("x86_64") != -1:
            current_host.CpuArch = "x86_64"
        elif properties.find("x86") != -1:
            current_host.CpuArch = "x86"
        elif properties.find("IA-64") != -1:
            current_host.CpuArch = "IA-64"
    
    current_host.CpuCount = int(info[1]["np"])
    current_host.TotalMemory = __calculateValue(attributes["physmem"])
    current_host.TotalSwap = __calculateValue(attributes["totmem"]) - current_host.TotalMemory
    current_host.CpuUtilization = float(attributes["loadave"])  

    jobs = attributes["jobs"]
    m = __jobsRegex.match(jobs)
    while m!=None:
        current_host.jobs.append(m.group(1))
        jobs = m.group(2)
        m = __jobsRegex.match(jobs)
    return current_host

def getJobData():
    global __jobsInfoList
    global __jobsInfoDict
    
    if __jobsInfoList is not None:
        return (__jobsInfoList,__jobsInfoDict)

    __jobsInfoDict = dict()
    __jobsInfoList = []
    qstat_output = _qstat(['-f', '-1'])
    job_list = __commandOutputToDict(qstat_output, None, __jobIdRegex)
    for jobid in job_list:
        job_info = job_list[jobid]
        job = __decodeJobData(jobid, job_info)
        if job != None:
            __jobsInfoDict[jobid] = job
            __jobsInfoList.append(job)


    return (__jobsInfoList,__jobsInfoDict)

def addExtendedJobData(job_ids, all_job_ids):
    jobInfos = asDict(getJobData())
    for job_id in job_ids:
        if job_id not in jobInfos:
            continue
        jobInfo = jobInfos[job_id]
        if jobInfo.JobStatus == "Pending" and jobInfo.EstStartTime is None:
            showstart_output = _showstart([jobInfo.JobID])
            if showstart_output is None:
                continue
            est_start_delta = __decodeShowstartTimeDelta(showstart_output)
            if est_start_delta is not None:
                jobInfo.EstStartTime = datetime.datetime.now() + est_start_delta




def __decodeJobData(jobid, attributes):
    current_job = JobInfo()
    current_job.JobID = jobid
    #guard agains weired output
    if "Job_Name" not in attributes:
      return None

    current_job.JobName = attributes["Job_Name"]

    if attributes["job_state"] == "R" or attributes["job_state"] == "E":
        current_job.JobStatus = "Running"
    elif attributes["job_state"] == "Q":
        current_job.JobStatus = "Pending"
    elif attributes["job_state"] == "C":
        current_job.JobStatus = "Done"
    else:
        current_job.JobStatus = "Unknown"
    current_job.OwnerName = attributes["Job_Owner"].split("@")[0]
    current_job.QueueName = attributes["queue"]
    current_job.SubmissionTime = datetime.datetime(*time.strptime(attributes["qtime"])[:6])
    if attributes["job_state"] == "Q":
        current_job.StartTime = None
    else:
        current_job.StartTime = datetime.datetime(*time.strptime(attributes["mtime"])[:6])
    hosts = {}
    if "exec_host" in attributes:
        for m in __execHostRegex.finditer(attributes["exec_host"]): 
            host_name = m.group(1)
            if host_name in hosts:
                hosts[host_name] += 1
            else:
                hosts[host_name]  = 1 

    if "resources_used.cput" in attributes:
      current_job.CpuTime = __decodeTimeDelta(attributes["resources_used.cput"])
    if "resources_used.walltime" in attributes:
      current_job.Walltime = __decodeTimeDelta(attributes["resources_used.walltime"])
    if "resources_used.mem" in attributes:
      current_job.MemUsage = int(__calculateValue(attributes["resources_used.mem"]) / (1024*1024)) #MemUsage is in Megabytes

    if "Resource_List.nodes" in attributes:
      current_job.TotalCores = __decodeNodes(attributes["Resource_List.nodes"])
    elif "Resource_List.procs" in attributes:
      current_job.TotalCores = int(attributes["Resource_List.procs"])
 
    if "submit_args" in attributes:
      current_job.Description = attributes["submit_args"]

    if ("submit_args" in attributes) and (attributes["submit_args"].find("FLAGS:ADVRES:") != -1):
      current_job.ReservationID =  __resIdRegex.search(attributes["submit_args"]).group(1)
    
    if ("x" in attributes) and (attributes["x"].find("FLAGS:ADVRES:") != -1):
      current_job.ReservationID =  __resIdRegex.search(attributes["x"]).group(1)

    for host in hosts:
        slots = hosts[host]
        current_job.ExecutionHosts[host] = slots
        if host in getHostData()[1] and getHostData()[1][host].State == "down":
            current_job.addComment("One of the job execution host (" + host + ") is down")
    #extra checks
    if current_job.JobStatus == "Pending" and \
       current_job.QueueName in getQueueData()[1] and \
       getQueueData()[1][current_job.QueueName].Started == False:
        current_job.addComment("Queue " + current_job.QueueName + " is not started")

    return current_job
    
    
def getQueueData():
    global __queuesInfoList
    global __queuesInfoDict
    if __queuesInfoList is not None:
       return (__queuesInfoList, __queuesInfoDict)

    qstat_output = _qstat([ '-f', '-Q', '-1'])
    all_queues = __commandOutputToDict(qstat_output, None, __queueRegex)

    __queuesInfoDict = dict() 
    __queuesInfoList = []
    for queue_name in all_queues:
            queue_dict = all_queues[queue_name]
            decodedQueue = __decodeQueueData(queue_name, queue_dict) 
            __queuesInfoDict[queue_name] = decodedQueue   
            __queuesInfoList.append(decodedQueue)   

    return (__queuesInfoList,__queuesInfoDict)

def __decodeQueueData(queue_name, attributes):
    current_queue = QueueInfo()
    current_queue.QueueName = queue_name
    queue_state = __decodeQueueStateField(attributes["state_count"])    
    current_queue.PendingJobs = int(queue_state["Queued"])  
    current_queue.RunningJobs = int(queue_state["Running"])
    current_queue.Enabled = (attributes["enabled"] == "True")
    current_queue.Started = (attributes["started"] == "True")
    if "resources_default.walltime" in attributes.keys():
        current_queue.Walltime = __decodeTimeDelta(attributes["resources_default.walltime"])
    else:
        current_queue.Walltime = datetime.timedelta(0)  
    current_queue.job_slots_aval = 0
    current_queue.job_slots_total = 0
    return current_queue

def _qstat(args):
    try:
        command = ['qstat']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError as e:
        raise GMExternalError("qstat command failed, reason:\n" + str(e))        
    
    qstat_output = sr.communicate()[0]
    if sr.returncode != 0:
        if sr.returncode == 153:
            raise GMUnknowJob("Unknow Job")
        else:
            raise GMExternalError("qstat command returned code: " + str(sr.returncode) + ". Details: " + qstat_output)
    return qstat_output


def _pbsnodes(args):
    try:
        command = ['pbsnodes']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError as e:
        raise GMExternalError("pbsnodes command failed, reason:\n" + str(e))        
    
    pbsnodes_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("pbsnodes command returned code: " + str(sr.returncode) + ". Details: " + pbsnodes_output)
    return pbsnodes_output

def _showres(args):
    try:
        command = ['showres']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
      # return dummy answer
      return "<response><data><sys time='1334223891'></sys></data><status code='0'>success</status></response>"
    
    showres_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("showres command returned code: " + str(sr.returncode))
    return showres_output

def _showstart(args):
    try:
        command = ['showstart']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        error("showstart command failed, reason:\n" + traceback.format_exc())
        return None
    
    showstart_output = sr.communicate()[0]
    if sr.returncode != 0:
        error("showstart command returned code: " + str(sr.returncode))
        return None

    return showstart_output


def _checknode(args):
    try:
        command = ['checknode']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("checknode command failed, reason:\n" + traceback.format_exc())       
    
    checknode_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("checknode command returned code: " + str(sr.returncode))
    return checknode_output

def _diagnose(args):
    try:
        command = ['diagnose']
        command.extend(args)
        sr = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise GMExternalError("diagnose command failed, reason:\n" + traceback.format_exc())        
    
    diagnose_output = sr.communicate()[0]
    if sr.returncode != 0:
        raise GMExternalError("diagnose command returned code: " + str(sr.returncode))
    return diagnose_output

def getReservationData():
    global __reservationsInfoList
    global __reservationsInfoDict
    if __reservationsInfoList is not None:
        return (__reservationsInfoList, __reservationsInfoDict)
    __reservationsInfoList = []
    __reservationsInfoDict = dict()
    showres_output = _showres(['-n', '--format=xml'])
    showres_output = showres_output.replace("Response>","response>").replace("Data>","data>")
    doc = libxml2.parseDoc(showres_output)
    ctx = doc.xpathNewContext()
    status_code = ctx.xpathEval("/response/status/@code")
    if len(status_code) == 1 and status_code[0].content != "0":
        raise GMInternalError("showres status code was: " + status_code[0].content)
    reservation_doc = ctx.xpathEval("/response/data/node")
    if len(reservation_doc) == 0:
       reservation_doc = ctx.xpathEval("/data/node") #Moab
    
    
    for node in reservation_doc:        
        ctx.setContextNode(node)
        host_name = ctx.xpathEval("@name")[0].content
        nres = ctx.xpathEval("nres")[0]
        ctx.setContextNode(nres)
        res_name = ctx.xpathEval("@Name")[0].content        
        
        res_type = ctx.xpathEval("@Type")[0].content
        if res_type == "User":
            resolved_res_type = "ar"
        elif res_type == "System":
            resolved_res_type = "sr"
        elif res_type == "Job":
            resolved_res_type = "job"
        else:
            raise GMInternalError("Reservation type: " + res_type + " is unknown")

        res_tc = int(ctx.xpathEval("@TC")[0].content)
 
        if(res_name not in __reservationsInfoDict):            
            __reservationsInfoDict[res_name] = ReservationInfo()
            __reservationsInfoDict[res_name].ReservationName = res_name                   
            __reservationsInfoList.append(__reservationsInfoDict[res_name])
        else:
            __reservationsInfoDict[res_name].Resources[host_name] = res_tc
            continue
        
        __reservationsInfoDict[res_name].Resources[host_name] = res_tc
        __reservationsInfoDict[res_name].Type = resolved_res_type
        
        start_time = ctx.xpathEval("@Start")[0].content
        end_time = ctx.xpathEval("@End")[0].content
        
        __reservationsInfoDict[res_name].StartTime = datetime.datetime.fromtimestamp(int(start_time))
        __reservationsInfoDict[res_name].EndTime = datetime.datetime.fromtimestamp(int(end_time))                     

    checknode_map = {}
    
    #help libmxl with memory mgmt
    doc.freeDoc()
    ctx.xpathFreeContext()
        
    return (__reservationsInfoList, __reservationsInfoDict)

