# $Id$

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from datetime import datetime, timedelta
import os

try:
    from qcg.core import *
    from qcg.comp import *
except ImportError:
    class Module:
        pass
    def error(msg):
        print("[ERROR]: " + msg)
    def info(msg):
        print("[INFO]: " + msg)
    def debug(msg):
        print("[DEBUG]: " + msg)


try:
    import grid_monitoring
    from grid_monitoring.gmexceptions import GMError, GMInternalError, GMExternalError, GMUnknowJob
    from grid_monitoring.common import asDict,asList
except ImportError:
    error("Unable to load grid_monitoring module")
    raise

from xml.dom.minidom import Document
import datetime
import sys

__MonitoringNSPrefix = "monitoring"
__MonitoringNS = "http://schemas.qoscosgrid.org/comp/2011/04/monitoring"
__QCGFactoryNSPrefix = "qcf"
__QCGFactoryNS = "http://schemas.qoscosgrid.org/comp/2011/04/factory"

__FactoryNS = "http://schemas.qoscosgrid.org/comp/2011/04/factory"
__BesFactoryNSPrefix = "bes-factory"
__BesFactoryNS = "http://schemas.ggf.org/bes/2006/08/bes-factory"
__JsdlNSPrefix = "jsdl"
__JsdlNS = "http://schemas.ggf.org/jsdl/2005/11/jsdl"

def __toXmlDuration(td):
    xmlPeriod = "PT"
    hours = td.days * 24
    hoursA = td.seconds / 3600
    hours = hours + hoursA
    sec = td.seconds - hoursA * 3600
    minutes = sec / 60
    sec = sec - minutes * 60
    xmlPeriod = xmlPeriod + str(hours) + "H" + str(minutes) + "M" + str(sec) + "S"
    return xmlPeriod

def __toXmlDateTime(dt):
    return dt.replace(microsecond=0).isoformat()

def jobMonitoringXmlFromJobInfo(jobInfo):    
    """Gets XML from JobInfo entity"""
    doc = Document()        
    activityInfo = doc.createElementNS(__QCGFactoryNS, "ActivityInfo")        
    activityInfo.setAttribute("xmlns", __QCGFactoryNS)
    doc.appendChild(activityInfo)

    if jobInfo.EstStartTime is not None:
        elem = doc.createElement("EstimatedStartTime")
        text = doc.createTextNode(__toXmlDateTime(jobInfo.EstStartTime))
        elem.appendChild(text)
        activityInfo.appendChild(elem)

    if jobInfo.StartTime is not None:
        elem = doc.createElement("StartTime")
        text = doc.createTextNode(__toXmlDateTime(jobInfo.StartTime))
        elem.appendChild(text)
        activityInfo.appendChild(elem)

    if jobInfo.EndTime is not None:
        elem = doc.createElement("EndTime")
        text = doc.createTextNode(__toXmlDateTime(jobInfo.EndTime))
        elem.appendChild(text)
        activityInfo.appendChild(elem)

    executionNodes = ""    
    for hostName, hostSlots in jobInfo.ExecutionHosts.items():
        if len(executionNodes) > 0:
            executionNodes = executionNodes + "|"
        if hostSlots is not None:
            executionNodes = executionNodes + hostName + ":" + str(hostSlots)
        else:
            executionNodes = executionNodes + hostName
    
    if len(executionNodes):
        elem = doc.createElement("ExecutionNodes")
        text = doc.createTextNode(executionNodes)
        elem.appendChild(text)
        activityInfo.appendChild(elem)

    if jobInfo.MemUsage is not None:
        elem = doc.createElement("MemoryUsageInMB")
        text = doc.createTextNode(str(jobInfo.MemUsage))
        elem.appendChild(text)
        activityInfo.appendChild(elem)

    if jobInfo.CpuTime is not None and \
       jobInfo.TotalCores is not None and \
       jobInfo.Walltime is not None:
        walltime = jobInfo.Walltime
        normWalltime = (walltime.days * 3600 * 24 + walltime.seconds) * jobInfo.TotalCores
        cpuTimeSeconds = jobInfo.CpuTime.days * 3600 * 24 + jobInfo.CpuTime.seconds
        elem = doc.createElement("Efficiency")
        if normWalltime > 0:
           text = doc.createTextNode(str(int(round((cpuTimeSeconds *100.0)/normWalltime))))
        else:
           text = doc.createTextNode("0")
        elem.appendChild(text)
        activityInfo.appendChild(elem)

    if jobInfo.Comment is not None:
        elem = doc.createElement("Comment")
        text = doc.createTextNode(jobInfo.Comment)
        elem.appendChild(text)
        activityInfo.appendChild(elem)


    return doc

def containedResourcesFromHostInfos(hostsInfo):
    """Gets XML from list of HostInfo entities"""
   
    doc = Document()

    cluster = doc.createElementNS(__BesFactoryNS, "ContainedResources")    
    cluster.setAttribute("xmlns:" + __BesFactoryNSPrefix, __BesFactoryNS)
    cluster.setAttribute("xmlns:" + __JsdlNSPrefix, __JsdlNS)
    doc.appendChild(cluster)        

    if len(asList(hostsInfo)) == 0:
        raise GMError("No hosts returned") 

    for host in asList(hostsInfo):
        containedResource = doc.createElement("ContainedResource")
        cluster.appendChild(containedResource)
        
        resourceName = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":ResourceName")
        containedResource.appendChild(resourceName)
        text = doc.createTextNode(str(host.HostName))
        resourceName.appendChild(text)
        
        if not host.CpuArch == "Unknown":                         
            cpuArchitecture = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":CPUArchitecture")
            containedResource.appendChild(cpuArchitecture)
            cpuArchitectureName = doc.createElementNS(__JsdlNS, __JsdlNSPrefix + ":CPUArchitectureName")
            cpuArchitecture.appendChild(cpuArchitectureName)
            text = doc.createTextNode(str(host.CpuArch))
            cpuArchitectureName.appendChild(text)
        
        cpuCount = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":CPUCount")
        containedResource.appendChild(cpuCount)
        text = doc.createTextNode(str(int(host.CpuCount)))
        cpuCount.appendChild(text)

        if(host.TotalMemory != None):
            physicalMemory = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":PhysicalMemory")
            containedResource.appendChild(physicalMemory)
            text = doc.createTextNode(str(int(host.TotalMemory)))
            physicalMemory.appendChild(text)
        
        if(host.TotalSwap != None):
            virtualMemory = doc.createElementNS(__BesFactoryNS, __BesFactoryNSPrefix + ":VirtualMemory")
            containedResource.appendChild(virtualMemory)
            text = doc.createTextNode(str(int(host.TotalSwap)))
            virtualMemory.appendChild(text)
        
        if(host.CpuUtilization != None):
            utilization = doc.createElementNS(__MonitoringNS, "Utilization")
            containedResource.appendChild(utilization)
            utilization.setAttribute("xmlns", __MonitoringNS)
            text = doc.createTextNode(str(int(host.CpuUtilization * 100)))
            utilization.appendChild(text)

        if(host.AdditionalResources != None):
            attrs = doc.createElementNS(__MonitoringNS, "Attributes")
            containedResource.appendChild(attrs)
            attrs.setAttribute("xmlns", __MonitoringNS)
            text = doc.createTextNode(str(host.AdditionalResources))
            attrs.appendChild(text)

        if(host.State != None):
            state = doc.createElementNS(__MonitoringNS, "State")
            containedResource.appendChild(state)
            state.setAttribute("xmlns", __MonitoringNS)
            text = doc.createTextNode(str(host.State))
            state.appendChild(text)
        if(host.Dedicated != None):
            dedicated = doc.createElementNS(__MonitoringNS, "Dedicated")
            containedResource.appendChild(dedicated)
            dedicated.setAttribute("xmlns", __MonitoringNS)
            text = doc.createTextNode(str(host.Dedicated).lower())
            dedicated.appendChild(text)
        if(host.GpuCount != None):
            gpucount = doc.createElementNS(__MonitoringNS, "GPUCount")
            containedResource.appendChild(gpucount)
            gpucount.setAttribute("xmlns", __MonitoringNS)
            text = doc.createTextNode(str(host.GpuCount).lower())
            gpucount.appendChild(text)


        usedCPUs = doc.createElementNS(__MonitoringNS, "UsedCPUCount")
        containedResource.appendChild(usedCPUs)
        usedCPUs.setAttribute("xmlns", __MonitoringNS)
        text = doc.createTextNode(str(int(host.UsedCpuCount)))
        usedCPUs.appendChild(text)
       
    return doc

def clusterMonitoringFromQueuesResAndJobs(queuesInfo, reservationsInfo, jobsInfo):
    """Gets XML from list of QueueInfo and Job entities"""
       
    doc = Document()
    
    cluster = doc.createElementNS(__MonitoringNS, "ClusterMonitoring")    
    cluster.setAttribute("xmlns", __MonitoringNS)
    doc.appendChild(cluster)
    
    queues = doc.createElementNS(__MonitoringNS, "Queues")
    cluster.appendChild(queues)
    if os.getenv("QCG_DEFAULT_QUEUE") is not None:
        default_queue = os.getenv("QCG_DEFAULT_QUEUE")
    else:
        default_queue = None
    
    for queueInfo in queuesInfo[0]:
        queue = doc.createElementNS(__MonitoringNS, "Queue")
        queue.setAttribute("name", queueInfo.QueueName)
        if queueInfo.Walltime == None:
            raise ValueError("Walltime must not be None")
        queue.setAttribute("walltime", __toXmlDuration(queueInfo.Walltime))
        queue.setAttribute("runningJobs", str(queueInfo.RunningJobs))
        queue.setAttribute("waitingJobs", str(queueInfo.PendingJobs))
        if queueInfo.Enabled is not None:
            queue.setAttribute("enabled", str(queueInfo.Enabled).lower())
        if queueInfo.Started is not None:
            queue.setAttribute("started", str(queueInfo.Started).lower())
        if default_queue is not None and default_queue == queueInfo.QueueName:
            queue.setAttribute("default", "true")
        queues.appendChild(queue) 
    
    jobs = doc.createElementNS(__MonitoringNS, "Jobs")
    cluster.appendChild(jobs)

    reservationsDict = dict()
    #parse reservations Info
    reservations = doc.createElementNS(__MonitoringNS, "Reservations")
    cluster.appendChild(reservations)
        
    for reservationInfo in reservationsInfo[0]:
        if reservationInfo.Type != "ar":
            reservationsDict[reservationInfo.ReservationName] = reservationInfo
            continue

        reservation = doc.createElementNS(__MonitoringNS, "Reservation")
        reservation.setAttribute("name", reservationInfo.ReservationName)
        reservation.setAttribute("type", reservationInfo.Type)
        if reservationInfo.StartTime == None:
            raise ValueError("StartTime must not be None")
        reservation.setAttribute("startTime", __toXmlDateTime(reservationInfo.StartTime))
        if reservationInfo.EndTime != None:
           reservation.setAttribute("endTime", __toXmlDateTime(reservationInfo.EndTime))
        if reservationInfo.Owner != None:
            reservation.setAttribute("owner", reservationInfo.Owner)

        reservations.appendChild(reservation)

        resources = reservationInfo.Resources
        for resourceKey in resources.keys():
            resource = doc.createElementNS(__MonitoringNS, "Host")
            resource.setAttribute("name", resourceKey)
            if resources[resourceKey] is not None:
                resource.setAttribute("slots", str(resources[resourceKey]))
            reservation.appendChild(resource)

        reservations.appendChild(reservation)

          
    for jobInfo in jobsInfo[0]:
        
        job = doc.createElementNS(__FactoryNS, "Job")
        job.setAttribute("name", jobInfo.JobName)
        job.setAttribute("id", jobInfo.JobID)
        job.setAttribute("queue", jobInfo.QueueName)
        if jobInfo.SubmissionTime != None:
            job.setAttribute("submissionTime", __toXmlDateTime(jobInfo.SubmissionTime))
        job.setAttribute("owner", jobInfo.OwnerName)

        if jobInfo.StartTime != None:
            job.setAttribute("startTime", __toXmlDateTime(jobInfo.StartTime))
        elif jobInfo.JobID in reservationsDict:
            job.setAttribute("estStartTime", __toXmlDateTime(reservationsDict[jobInfo.JobID].StartTime))

        if jobInfo.EndTime != None:
            job.setAttribute("endTime", __toXmlDateTime(jobInfo.EndTime))
        elif jobInfo.JobID in reservationsDict:
            job.setAttribute("estEndTime", __toXmlDateTime(reservationsDict[jobInfo.JobID].EndTime))

        if jobInfo.TotalCores != None:
            job.setAttribute("cores", str(jobInfo.TotalCores))

        if jobInfo.CpuTime != None:
            job.setAttribute("cpuTime", str(jobInfo.CpuTime))

        if jobInfo.MemUsage != None:
            job.setAttribute("memUsage", str(jobInfo.MemUsage))

        if jobInfo.Description != None:
            job.setAttribute("desc", jobInfo.Description)

        if jobInfo.ReservationID != None:
            job.setAttribute("resId", jobInfo.ReservationID)

        if len(jobInfo.ExecutionHosts) != 0:
            eHosts = jobInfo.ExecutionHosts
            for ehost in eHosts.keys():
                host = doc.createElementNS(__MonitoringNS, "Host")
                host.setAttribute("name", ehost)
                if eHosts[ehost] is not None:
                    host.setAttribute("slots", str(eHosts[ehost]))
                job.appendChild(host)
        elif jobInfo.JobID in reservationsDict:
            rHosts = reservationsDict[jobInfo.JobID].Resources
            for rhost in rHosts.keys():
                host = doc.createElementNS(__MonitoringNS, "Host")
                host.setAttribute("name", rhost)
                if rHosts[rhost] is not None:
                    host.setAttribute("slots", str(rHosts[rhost]))
                job.appendChild(host)
        jobs.appendChild(job)
    return doc

def removeXmlHeader(string):
    return string.replace("<?xml version=\"1.0\" ?>", "")


_cap_map = grid_monitoring.monitoringCapabilities()

class MonitoringDRMS(Module):
    
    def __init__(self, *unused):
        signal_connect("smc_factory_attributes", self.factory_attr_handler)
        signal_connect("smc_contained_resources", self.contained_resources_handler)
        signal_connect("smc_jobs_info", self.jobs_info_handler)

    def factory_attr_handler(self, signal, drms_type, unused):
        try:
            queuesInfo = grid_monitoring.getQueueData(drms_type)
            try:
                reservationsInfo = grid_monitoring.getReservationData(drms_type)
            except:
                error("getReservationData failed: " + str(sys.exc_info()[0]))
                reservationsInfo = ([], dict()) #continue - provding info about reservations is not so crucial
            
            jobsInfo = grid_monitoring.getJobData(drms_type)
            doc = clusterMonitoringFromQueuesResAndJobs(queuesInfo, reservationsInfo, jobsInfo)
            result = removeXmlHeader(doc.toxml())
            doc.unlink()
            return result
        except GMError as e:
            error("Failed to get factory attr: " + str(e))
            signal.stop_emission();
    
    def contained_resources_handler(self, signal, drms_type, unused):
        try:
            hostsInfo = grid_monitoring.getHostData(drms_type)
            doc = containedResourcesFromHostInfos(hostsInfo)                  
            result = removeXmlHeader(doc.toxml())
            doc.unlink()
            return result
        except GMError as e:
            error("Failed to get contained resources attr: " + str(e))
            signal.stop_emission();
    
    def jobs_info_handler(self, signal, drms_type, job_ids, extended, unused):
        try:
            debug("job_ids = " + str(job_ids) + " extended = " + str(extended))
            #get basic data for all jobs (this information is cached so it costs nothing)
            jobInfos = grid_monitoring.getJobData(drms_type)
            #construct subset of jobs you we are especially interested in
            job_ids_subset = []
            users_jobs_count_map = dict()
            for jobInfo in asList(jobInfos): #we are iterating over jobInfos not jobId beacuse jobInfos should maintain job priority
                if jobInfo.JobID not in job_ids:
                   continue
                #job submitted within last 5-15m
                if jobInfo.SubmissionTime is not None:
                     since_submission = datetime.datetime.now() - jobInfo.SubmissionTime
                else:
                     since_submission = None

                if since_submission is not None and \
                     since_submission  >  timedelta(minutes=5) and \
                     since_submission  <  timedelta(minutes=15):
                     pass
                elif not extended: #this litmits frequency
                    continue

                if jobInfo.JobStatus != "Pending": #TODO in the future we may need to keep separate list for running/pending jobs
                    continue                 

                #golden fish can tell you only about 3 jobs
                if jobInfo.OwnerName in users_jobs_count_map:
                    if users_jobs_count_map[jobInfo.OwnerName] >= 3:
                        continue
                    users_jobs_count_map[jobInfo.OwnerName] = users_jobs_count_map[jobInfo.OwnerName] + 1  
                else:
                    users_jobs_count_map[jobInfo.OwnerName] = 1
                #and the lucky job_id is...
                job_ids_subset.append(jobInfo.JobID)

            #safet valve
            job_ids_subset = job_ids_subset[:30]
            info("Jobs selected for extended query: " + str(job_ids_subset))

            try:
                grid_monitoring.addExtendedJobData(drms_type, job_ids_subset, job_ids)
            except: #be on safe side
                error("Failed to get extended job information")
          
            #create xml documents, one for each job
            responses = []
            for job_id in job_ids:
                if job_id in asDict(jobInfos):
                    doc = jobMonitoringXmlFromJobInfo(asDict(jobInfos)[job_id])
                    responses.append(removeXmlHeader(doc.toxml()))
                    doc.unlink()
                else: # job may be terminated already
                    debug("Job " + job_id + " already terminated")
                    responses.append("")

            return [responses]
        except GMUnknowJob:
            return ""


if __name__ == '__main__':

    def signal_connect(*args):
      pass

    def debug(*args):
      pass

    def info(*args):
      pass

    def error(*args):
      pass

    monitoring = MonitoringDRMS()

    drms_type="http://schemas.qoscosgrid.org/comp/2011/04/drms/PBSPro/"
    #drms_type="http://schemas.qoscosgrid.org/comp/2011/04/drms/Torque/Torque"
    #drms_type="http://schemas.qoscosgrid.org/comp/2011/04/drms/SLURM/"


    factory_attr = monitoring.factory_attr_handler(signal=None, drms_type=drms_type, unused=None)
    resources = monitoring.contained_resources_handler(signal=None, drms_type=drms_type, unused=None)
    jobs = monitoring.jobs_info_handler(signal=None, drms_type=drms_type, job_ids=[], extended=[], unused=None)

    print("<doc_factory>")
    print("<factory_attr>%s</factory_attr>" % factory_attr)
    print("<resources>%s</resources>" % resources)

    if jobs and jobs[0]:
       print("<jobs>%s</jobs>" % jobs)

    print("</doc_factory>")
