# $Id: reservation_sge.py 2588 2009-07-28 11:03:09Z mmamonski $
#
# vim: set ts=4 sw=4 et:

#
# QosCosGrid Computing
# Copyright (C) 2008-2011 Poznan Supercomputing and Networking Center
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

try:
  from lxml import etree
except ImportError:
  try:
    # Python 2.5
    import xml.etree.cElementTree as etree
  except ImportError:
    try:
      # Python 2.5
      import xml.etree.ElementTree as etree
    except ImportError:
      try:
        # normal cElementTree install
        import cElementTree as etree
      except ImportError:
        try:
          # normal ElementTree install
          import elementtree.ElementTree as etree
        except ImportError:
          error("Failed to import ElementTree from any known place")
          raise

from commands import getstatusoutput
import re
import os
from qcg.core import *
from qcg.comp import *

ns = "{" + SMC_NAMESPACE_ARDL + "}"

class InternalFaultException(Exception):
    pass

class NotAuthorizedException(Exception):
    pass

class UnsupportedFeatureException(Exception):
    pass

class InvalidRequestMessageException(Exception):
    pass

class UnknownReservationIdentifierException(Exception):
    pass

class NoSlotsAvailableException(Exception):
    pass

class NoMatchinResourcesException(Exception):
    pass

class ReservationMaui(Module):

    def convertTime(self, dt):
        from calendar import timegm
        from time import strptime, strftime, localtime

        # The XML dateTime value is always in UTC and with no fractions because it is
        # serialized and then deserialized by gSOAP which uses this constant format.
        utc_tuple = strptime(dt, "%Y-%m-%dT%H:%M:%SZ")
        utc_seconds = timegm(utc_tuple)
        local_tuple = localtime(utc_seconds)
        # e.g.: 21:00:00_09/3/09 
        return strftime("%H:%M:%S_%m/%d/%y", local_tuple)

    def create(self, ardl):
        try:
            res = self.internal_create(ardl)
            return (RetCode(RESERVATION_OK), res)
        except NotAuthorizedException:
            return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, "Not authorized")
        except InvalidRequestMessageException, e:
            return RetCode(RESERVATION_INVALID_REQUEST_MESSAGE_FAULT, str(e) or repr(e))
        except NoSlotsAvailableException:
            return RetCode(RESERVATION_NO_SLOTS_AVAILABLE_FAULT, "No slots available")
        except NoMatchinResourcesException:
            return RetCode(RESERVATION_NO_MATCHING_RESOURCES_FAULT, "No matching resources")
        except UnsupportedFeatureException, e:
            return RetCode(RESERVATION_UNSUPPORTED_FEATURE_FAULT, str(e))
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def terminate(self, arid):
        try:
            self.internal_terminate(arid)
            return RetCode(RESERVATION_OK)
        except UnknownReservationIdentifierException:
            return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
        except NotAuthorizedException, e:
            return RetCode(RESERVATION_NOT_AUTHORIZED_FAULT, str(e) or repr(e))
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def get_active(self):
        try:
            res = self.internal_get_active()
            return [RetCode(RESERVATION_OK)] + res
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def get_status(self, arid):
        try:
            res = self.internal_get_status(arid)
            return (RetCode(RESERVATION_OK), res)
        except UnknownReservationIdentifierException:
            return RetCode(RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT, arid)
        except Exception, e:
            return RetCode(RESERVATION_INTERNAL_FAULT, str(e) or repr(e))

    def internal_create(self, ardl):
        parsed = etree.fromstring(ardl)
        descr = parsed.find(ns + "ReservationDescription")
        resources = descr.find(ns + "Resources")

        # TODO
        unsupported = ("FileSystem", "OperatingSystem", "CPUArchitecture",
         "NativeResource")

        for u in unsupported:
           if resources.find(ns + u) is not None:
               raise UnsupportedFeatureException(u)

        if resources.find(ns + "ExclusiveExecution") is not None:
            if resources.find(ns + "ExclusiveExecution").text == "true":
               raise UnsupportedFeatureException("ExclusiveExecution")

        cmd = "setres -x BYNAME -r PROCS=1"
   
        try:
            cmd += ":MEM=" + resources.find(ns + "IndividualPhysicalMemory").text
        except:
            pass
    
        try:
            cmd += ":SWAP=" + resources.find(ns + "IndividualSwapMemory").text
        except:
            pass
    
        cmd += " "

        try:
            cmd += " -n " + descr.find(ns + "ReservationIdentification").find(ns + "ReservationName").text
        except:
            pass
 
        if os.getenv("QCG_AR_MAUI_PARTITION") != None:
	    cmd += " -p " +  os.getenv("QCG_AR_MAUI_PARTITION") + " "
    
        slotsCount = 1
        try:
                slotsCount = resources.find(ns + "ReservedSlotsCount").text
        except:
                pass
    
        additionalSlotsCount = 0
        try:
            additionalSlotsCount = resources.find(ns + "AdditionalReservedSlotsCount").text
        except:
            pass
    
        try:
            cmd += " -u " + resources.find(ns + "UserName").text
        except:
            pass
    
        try:
            cmd += " -g " + resources.find(ns + "GroupName").text
        except:
            pass
    
        try:
           candidateHosts = resources.find(ns + "CandidateHosts")
    
           if candidateHosts:
               #cmd += " -f "
               firstRun = True
               
               for hostName in candidateHosts.findall(ns + "HostName"):
                   if not firstRun:
                       raise UnsupportedFeatureException("Multiple HostName selection is currently unsupported")
               else: 
                   firstRun = False
    
               #cmd += hostName.text
    
        except:
            pass

        try:
            timeWindow = descr.find(ns + "TimeWindow")

            openWindow = timeWindow.get("open")
            if openWindow is not None and openWindow == "true":
                raise UnsupportedFeatureException("Open TimeWindow")

            startTime = timeWindow.find(ns + "StartTime").text
            debug("StartTime " + startTime)

            endTime = timeWindow.find(ns + "EndTime").text
            debug("EndTime " + endTime)

            startTime = self.convertTime(startTime)
            endTime = self.convertTime(endTime)

            cmd += " -s " + startTime + " -e " + endTime 
        except UnsupportedFeatureException, e:
            raise e
        except Exception, e:
            raise InvalidRequestMessageException("No StartTime or EndTime: " + str(e) or repr(e))

        #best effort semantic.
        output = ""
        for slots in range((int(additionalSlotsCount) + int(slotsCount)), (int(slotsCount) - 1), -1):
           
            command = cmd + " TASKS==" + str(slots)

            info("cmd = " + command)

            output = getstatusoutput(command)
            
            if output[0] == 0:
                return output[1].splitlines()[3].split()[1].strip('\'')

        if re.search("cannot select", output[1]):
            raise NoSlotsAvailableException()
        if re.search("not authorized", output[1]):
            raise NotAuthorizedException()

        notice("setres failed: " + output[1])

        raise InternalFaultException(output[1])

    def internal_get_data(self, command):
        output = getstatusoutput(command)

        if output[0] != 0:
            raise InternalFaultException(output[1])

        return output[1].splitlines()

    def internal_get_active(self):
        lines = self.internal_get_data("showres -g")
        # Return only User resrvation  (not the 'Job' reservations). TODO what about Group reservations?
        return [line.split()[0] for line in lines if line.find('User') != -1]

    def internal_get_status(self, arid):
        lines = self.internal_get_data("showres -n -g " + arid);    

        hostsformatted = []
    
        for line in lines:
            if line.find(arid) != -1:
                host = line.split()[0]
                hlines = self.internal_get_data("checknode " + host);
                procsCount = re.search("PROCS:\s+([0-9]+)", ''.join(hlines)).group(1)
                
                for i in range(len(hlines)):
                    if hlines[i].find("'" + arid + "'" ) != -1:
                        reservedProcs = re.search("Procs:\s+([0-9]+)/",hlines[i+1]).group(1)
                        hostsformatted.append(host + ":" + procsCount + ":" +  reservedProcs + ":0")
    
        assert len(hostsformatted) != 0
    
            # TODO: get slots used by jobs within this reservation. Note -- data may be inconsistent
            # because more than one call to MAUI was needed
        return hostsformatted

    def internal_terminate(self, arid):
        res = getstatusoutput("releaseres " + arid)

    # exit code means nothing....
        if re.search("released", res[1]):
            return

        if re.search("cannot locate reservation", res[1]):
            raise UnknownReservationIdentifierException
    
        if re.search("not authorized". res[1]):
            raise NotAuthorizedException

        notice("releaseres failed: " + res[1])
        
        raise InternalFaultException(res[1])
    
